@@ -23,43 +23,51 @@
 // License along with this library; if not, write to the Free Software
 // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-
 using System;
 using System.Data;
-using System.Net;
-using System.Net.Sockets;
-using System.IO;
 using System.Text;
+using System.Resources;
 using System.ComponentModel;
 using System.Collections;
+
 using NpgsqlTypes;
+
+using System.Text.RegularExpressions;
+
+#if WITHDESIGN
 using Npgsql.Design;
+#endif
 
 namespace Npgsql
 {
     /// <summary>
-    /// Represents a SQL statement or function (stored procedure) to execute against a PostgreSQL database. This class cannot be inherited.
+    /// Represents a SQL statement or function (stored procedure) to execute
+    /// against a PostgreSQL database. This class cannot be inherited.
     /// </summary>
+    #if WITHDESIGN
     [System.Drawing.ToolboxBitmapAttribute(typeof(NpgsqlCommand)), ToolboxItem(true)]
-    public sealed class NpgsqlCommand : Component, IDbCommand
+    #endif
+    public sealed class NpgsqlCommand : Component, IDbCommand, ICloneable
     {
+        // Logging related values
+        private static readonly String CLASSNAME = "NpgsqlCommand";
+        private static ResourceManager resman = new ResourceManager(typeof(NpgsqlCommand));
 
         private NpgsqlConnection            connection;
+        private NpgsqlConnector             connector;
         private NpgsqlTransaction           transaction;
         private String                      text;
         private Int32                       timeout;
         private CommandType                 type;
         private NpgsqlParameterCollection   parameters;
         private String                      planName;
-        private static Int32                planIndex = 0;
-        private static Int32                portalIndex = 0;
 
         private NpgsqlParse                 parse;
         private NpgsqlBind                  bind;
 
-        // Logging related values
-        private static readonly String CLASSNAME = "NpgsqlCommand";
-        private System.Resources.ResourceManager resman;
+        private Boolean						invalidTransactionDetected = false;
+        
+        private CommandBehavior             commandBehavior;
 
         // Constructors
 
@@ -89,26 +97,42 @@ namespace Npgsql
         /// <param name="transaction">The <see cref="Npgsql.NpgsqlTransaction">NpgsqlTransaction</see> in which the <see cref="Npgsql.NpgsqlCommand">NpgsqlCommand</see> executes.</param>
         public NpgsqlCommand(String cmdText, NpgsqlConnection connection, NpgsqlTransaction transaction)
         {
-            resman = new System.Resources.ResourceManager(this.GetType());
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, CLASSNAME);
 
             planName = String.Empty;
             text = cmdText;
             this.connection = connection;
+            
+            if (this.connection != null)
+                this.connector = connection.Connector;
+
             parameters = new NpgsqlParameterCollection();
             timeout = 20;
             type = CommandType.Text;
             this.Transaction = transaction;
+            commandBehavior = CommandBehavior.Default;
+            
+            
         }
 
-        /*
         /// <summary>
-        /// Finalizer for <see cref="Npgsql.NpgsqlCommand">NpgsqlCommand</see>.
+        /// Used to execute internal commands.
         /// </summary>
-        ~NpgsqlCommand ()
+        internal NpgsqlCommand(String cmdText, NpgsqlConnector connector)
         {
-            Dispose(false);
-        }*/
+            resman = new System.Resources.ResourceManager(this.GetType());
+            NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, CLASSNAME);
+            
+
+            planName = String.Empty;
+            text = cmdText;
+            this.connector = connector;
+            type = CommandType.Text;
+            commandBehavior = CommandBehavior.Default;
+            
+            parameters = new NpgsqlParameterCollection();
+            timeout = 20;
+        }
 
         // Public properties.
         /// <summary>
@@ -130,6 +154,7 @@ namespace Npgsql
                 planName = String.Empty;
                 parse = null;
                 bind = null;
+                commandBehavior = CommandBehavior.Default;
             }
         }
 
@@ -175,7 +200,8 @@ namespace Npgsql
             }
         }
 
-        IDbConnection IDbCommand.Connection {
+        IDbConnection IDbCommand.Connection 
+        {
             get
             {
                 return Connection;
@@ -183,7 +209,7 @@ namespace Npgsql
 
             set
             {
-                connection = (NpgsqlConnection) value;
+                Connection = (NpgsqlConnection) value;
                 NpgsqlEventLog.LogPropertySet(LogLevel.Debug, CLASSNAME, "IDbCommand.Connection", value);
             }
         }
@@ -203,15 +229,35 @@ namespace Npgsql
 
             set
             {
-                if (this.transaction != null && this.transaction.Connection == null)
-                    this.transaction = null;
-                if (this.connection != null && this.connection.InTransaction == true)
+                if (this.Connection == value)
+                    return;
+
+                //if (this.transaction != null && this.transaction.Connection == null)
+                  //  this.transaction = null;
+                                    
+                if (this.transaction != null && this.connection != null && this.Connector.Transaction != null)
                     throw new InvalidOperationException(resman.GetString("Exception_SetConnectionInTransaction"));
+
+
                 this.connection = value;
+                Transaction = null;
+                if (this.connection != null)
+                    connector = this.connection.Connector;
+
                 NpgsqlEventLog.LogPropertySet(LogLevel.Debug, CLASSNAME, "Connection", value);
             }
         }
 
+        internal NpgsqlConnector Connector {
+            get
+            {
+                if (connector == null && this.connection != null)
+                    connector = this.connection.Connector;
+
+                return connector;
+            }
+        }
+
         IDataParameterCollection IDbCommand.Parameters {
             get
             {
@@ -223,7 +269,10 @@ namespace Npgsql
         /// Gets the <see cref="Npgsql.NpgsqlParameterCollection">NpgsqlParameterCollection</see>.
         /// </summary>
         /// <value>The parameters of the SQL statement or function (stored procedure). The default is an empty collection.</value>
+        #if WITHDESIGN
         [Category("Data"), DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
+        #endif
+        
         public NpgsqlParameterCollection Parameters {
             get
             {
@@ -232,14 +281,32 @@ namespace Npgsql
             }
         }
 
+        
+        IDbTransaction IDbCommand.Transaction 
+        {
+            get
+            {
+                return Transaction;
+            }
+
+            set
+            {
+                Transaction = (NpgsqlTransaction) value;
+                NpgsqlEventLog.LogPropertySet(LogLevel.Debug, CLASSNAME, "IDbCommand.Transaction", value);
+            }
+        }
+        
         /// <summary>
         /// Gets or sets the <see cref="Npgsql.NpgsqlTransaction">NpgsqlTransaction</see>
         /// within which the <see cref="Npgsql.NpgsqlCommand">NpgsqlCommand</see> executes.
         /// </summary>
         /// <value>The <see cref="Npgsql.NpgsqlTransaction">NpgsqlTransaction</see>.
         /// The default value is a null reference.</value>
+        #if WITHDESIGN
         [Browsable(false), DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
-        public IDbTransaction Transaction {
+        #endif
+        
+        public NpgsqlTransaction Transaction {
             get
             {
                 NpgsqlEventLog.LogPropertyGet(LogLevel.Debug, CLASSNAME, "Transaction");
@@ -265,7 +332,10 @@ namespace Npgsql
         /// method of the <see cref="System.Data.Common.DbDataAdapter">DbDataAdapter</see>.
         /// </summary>
         /// <value>One of the <see cref="System.Data.UpdateRowSource">UpdateRowSource</see> values.</value>
+        #if WITHDESIGN
         [Category("Behavior"), DefaultValue(UpdateRowSource.Both)]
+        #endif
+        
         public UpdateRowSource UpdatedRowSource {
             get
             {
@@ -292,6 +362,26 @@ namespace Npgsql
             // [TODO] Finish method implementation.
             throw new NotImplementedException();
         }
+        
+        /// <summary>
+        /// Create a new command based on this one.
+        /// </summary>
+        /// <returns>A new NpgsqlCommand object.</returns>
+        Object ICloneable.Clone()
+        {
+            return Clone();
+        }
+
+        /// <summary>
+        /// Create a new connection based on this one.
+        /// </summary>
+        /// <returns>A new NpgsqlConnection object.</returns>
+        public NpgsqlCommand Clone()
+        {
+            // TODO: Add consistency checks.
+
+            return new NpgsqlCommand(CommandText, Connection, Transaction);
+        }
 
         /// <summary>
         /// Creates a new instance of an <see cref="System.Data.IDbDataParameter">IDbDataParameter</see> object.
@@ -324,14 +414,18 @@ namespace Npgsql
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "ExecuteNonQuery");
 
             ExecuteCommand();
-
+            
+            UpdateOutputParameters();
+            
+            
             // If nothing is returned, just return -1.
-            if(connection.Mediator.CompletedResponses.Count == 0) {
+            if(Connector.Mediator.CompletedResponses.Count == 0)
+            {
                 return -1;
             }
 
             // Check if the response is available.
-            String firstCompletedResponse = (String)connection.Mediator.CompletedResponses[0];
+            String firstCompletedResponse = (String)Connector.Mediator.CompletedResponses[0];
 
             if (firstCompletedResponse == null)
                 return -1;
@@ -354,6 +448,80 @@ namespace Npgsql
             else
                 return -1;
         }
+        
+        
+        
+        private void UpdateOutputParameters()
+        {
+            // Check if there was some resultset returned. If so, put the result in output parameters.
+            NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "UpdateOutputParameters");
+            
+            // Get ResultSets.
+            ArrayList resultSets = Connector.Mediator.ResultSets;
+            
+            if (resultSets.Count != 0)
+            {
+                NpgsqlResultSet nrs = (NpgsqlResultSet)resultSets[0];
+                                
+                if ((nrs != null) && (nrs.Count > 0))
+                {
+                    NpgsqlAsciiRow nar = (NpgsqlAsciiRow)nrs[0];
+                    
+                    Int32 i = 0;
+                    Boolean hasMapping = false;
+                                        
+                    // First check if there is any mapping between parameter name and resultset name.
+                    // If so, just update output parameters which has mapping.
+                    
+                    foreach (NpgsqlParameter p in Parameters)
+                    {
+                        try
+                        {
+                            if (nrs.RowDescription.FieldIndex(p.ParameterName.Substring(1)) > -1)
+                            {
+                                hasMapping = true;
+                                break;
+                            }
+                        }
+                        catch(ArgumentOutOfRangeException)
+                        {}
+                    }
+                                        
+                    
+                    if (hasMapping)
+                    {
+                        foreach (NpgsqlParameter p in Parameters)
+                        {
+                            if (((p.Direction == ParameterDirection.Output) ||
+                                (p.Direction == ParameterDirection.InputOutput)) && (i < nrs.RowDescription.NumFields ))
+                            {
+                                try
+                                {
+                                    p.Value = nar[nrs.RowDescription.FieldIndex(p.ParameterName.Substring(1))];
+                                    i++;
+                                }
+                                catch(ArgumentOutOfRangeException)
+                                {}
+                            }
+                        }
+                        
+                    }
+                    else
+                        foreach (NpgsqlParameter p in Parameters)
+                        {
+                            if (((p.Direction == ParameterDirection.Output) ||
+                                (p.Direction == ParameterDirection.InputOutput)) && (i < nrs.RowDescription.NumFields ))
+                            {
+                                p.Value = nar[i];
+                                i++;
+                            }
+                        }
+                }
+                
+            }   
+            
+            
+        }
 
         /// <summary>
         /// Sends the <see cref="Npgsql.NpgsqlCommand.CommandText">CommandText</see> to
@@ -410,11 +578,14 @@ namespace Npgsql
             // [FIXME] No command behavior handling.
 
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "ExecuteReader", cb);
+            commandBehavior = cb;
 
             ExecuteCommand();
-
+            
+            UpdateOutputParameters();
+            
             // Get the resultsets and create a Datareader with them.
-            return new NpgsqlDataReader(connection.Mediator.ResultSets, connection.Mediator.CompletedResponses, connection, cb);
+            return new NpgsqlDataReader(Connector.Mediator.ResultSets, Connector.Mediator.CompletedResponses, cb, this);
         }
 
         ///<summary>
@@ -430,16 +601,18 @@ namespace Npgsql
                 for (Int32 i = 0; i < parameters.Count; i++)
                 {
                     // Do not quote strings, or escape existing quotes - this will be handled by the backend.
-                    parameterValues[i] = NpgsqlTypesHelper.ConvertNpgsqlParameterToBackendStringValue(parameters[i], false);
+                    // DBNull or null values are returned as null.
+                    // TODO: Would it be better to remove this null special handling out of ConvertToBackend??
+                    parameterValues[i] = parameters[i].TypeInfo.ConvertToBackend(parameters[i].Value, true);
                 }
                 bind.ParameterValues = parameterValues;
             }
 
-            connection.Bind(bind);
-            connection.Mediator.RequireReadyForQuery = false;
-            connection.Flush();
+            Connector.Bind(bind);
+            Connector.Mediator.RequireReadyForQuery = false;
+            Connector.Flush();
 
-            connection.CheckErrorsAndNotifications();
+            connector.CheckErrorsAndNotifications();
         }
 
         /// <summary>
@@ -470,7 +643,7 @@ namespace Npgsql
             // Only the first column of the first row must be returned.
 
             // Get ResultSets.
-            ArrayList resultSets = connection.Mediator.ResultSets;
+            ArrayList resultSets = Connector.Mediator.ResultSets;
 
             // First data is the RowDescription object.
             // Check all resultsets as insert commands could have been sent along
@@ -501,31 +674,33 @@ namespace Npgsql
             // Check the connection state.
             CheckConnectionState();
 
-            if (! connection.SupportsPrepare) {
+            if (! Connector.SupportsPrepare)
+            {
                 return;	// Do nothing.
             }
 
-            if (connection.BackendProtocolVersion == ProtocolVersion.Version2)
+            if (connector.BackendProtocolVersion == ProtocolVersion.Version2)
             {
-                NpgsqlCommand command = new NpgsqlCommand(GetPrepareCommandText(), connection );
+                NpgsqlCommand command = new NpgsqlCommand(GetPrepareCommandText(), connector );
                 command.ExecuteNonQuery();
             }
             else
             {
                 // Use the extended query parsing...
-                planName = "NpgsqlPlan" + System.Threading.Interlocked.Increment(ref planIndex);
-                String portalName = "NpgsqlPortal" + System.Threading.Interlocked.Increment(ref portalIndex);
+                //planName = "NpgsqlPlan" + Connector.NextPlanIndex();
+                planName = Connector.NextPlanName();
+                String portalName = Connector.NextPortalName();
 
                 parse = new NpgsqlParse(planName, GetParseCommandText(), new Int32[] {});
 
-                connection.Parse(parse);
-                connection.Mediator.RequireReadyForQuery = false;
-                connection.Flush();
+                Connector.Parse(parse);
+                Connector.Mediator.RequireReadyForQuery = false;
+                Connector.Flush();
 
                 // Check for errors and/or notifications and do the Right Thing.
-                connection.CheckErrorsAndNotifications();
+                connector.CheckErrorsAndNotifications();
 
-                bind = new NpgsqlBind(portalName, planName, new Int16[] {0}, null, new Int16[] {0});
+                bind = new NpgsqlBind("", planName, new Int16[] {0}, null, new Int16[] {0});
             }
         }
 
@@ -559,12 +734,12 @@ namespace Npgsql
         {
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "CheckConnectionState");
 
-            // Check the connection state.
-            if (connection == null)
-                throw new InvalidOperationException(resman.GetString("Exception_ConnectionNull"));
-            if (connection.State != ConnectionState.Open)
-                throw new InvalidOperationException(resman.GetString("Exception_ConnectionNotOpen"));
 
+            // Check the connection state.
+            if (Connector == null || Connector.State != ConnectionState.Open)
+            {
+                throw new InvalidOperationException(resman.GetString("Exception_ConnectionNotOpen"));
+            }
         }
 
         /// <summary>
@@ -586,37 +761,259 @@ namespace Npgsql
 
         private String GetClearCommandText()
         {
-            NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "GetClearCommandText");
+            if (NpgsqlEventLog.Level == LogLevel.Debug)
+                NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "GetClearCommandText");
 
+            Boolean addProcedureParenthesis = false;    // Do not add procedure parenthesis by default.
+
+            Boolean functionReturnsRecord = false;      // Functions don't return record by default.
+
+            Boolean functionReturnsRefcursor = false;   // Functions don't return refcursor by default.
 
             String result = text;
 
             if (type == CommandType.StoredProcedure)
-                if (connection.SupportsPrepare)
+            {
+
+                if (Parameters.Count > 0)
+                    functionReturnsRecord = CheckFunctionReturn("record");
+
+                functionReturnsRefcursor = CheckFunctionReturn("refcursor");
+
+                // Check if just procedure name was passed. If so, does not replace parameter names and just pass parameter values in order they were added in parameters collection. Also check if command text finishes in a ";" which would make Npgsql incorrectly append a "()" when executing this command text.
+                if ((!result.Trim().EndsWith(")")) && (!result.Trim().EndsWith(";")))
+                {
+                    addProcedureParenthesis = true;
+                    result += "(";
+                }
+
+                if (Connector.SupportsPrepare)
                     result = "select * from " + result; // This syntax is only available in 7.3+ as well SupportsPrepare.
                 else
-                    result = "select " + result;				// Only a single result return supported. 7.2 and earlier.
+                    result = "select " + result;        //Only a single result return supported. 7.2 and earlier.
+            }
             else if (type == CommandType.TableDirect)
-                return "select * from " + result; // There is no parameter support on table direct.
+                return "select * from " + result;       // There is no parameter support on table direct.
 
-            if (parameters.Count == 0)
-                return result;
-
-
-            //CheckParameters();
-
-            String parameterName;
-
-            for (Int32 i = 0; i < parameters.Count; i++)
+            if (parameters == null || parameters.Count == 0)
             {
-                parameterName = parameters[i].ParameterName;
+                if (addProcedureParenthesis)
+                    result += ")";
 
-                result = ReplaceParameterValue(result, parameterName, NpgsqlTypesHelper.ConvertNpgsqlParameterToBackendStringValue(parameters[i], true));
 
+                // If function returns ref cursor just process refcursor-result function call
+                // and return command which will be used to return data from refcursor.
+
+                if (functionReturnsRefcursor)
+                    return ProcessRefcursorFunctionReturn(result);
+
+
+                if (functionReturnsRecord)
+                    result = AddFunctionReturnsRecordSupport(result);
+
+
+                result = AddSingleRowBehaviorSupport(result);
+
+                return result;
             }
 
-            return result;
 
+            // Get parameters in query string to translate them to their actual values.
+
+            // This regular expression gets all the parameters in format :param or @param
+            // and everythingelse.
+            // This is only needed if query string has parameters. Else, just append the
+            // parameter values in order they were put in parameter collection.
+
+
+            // If parenthesis don't need to be added, they were added by user with parameter names. Replace them.
+            if (!addProcedureParenthesis)
+            {
+
+                Regex a = new Regex(@"(:[\w]*)|(@[\w]*)|(.)", RegexOptions.Singleline);
+
+                //CheckParameters();
+
+                StringBuilder sb = new StringBuilder();
+
+                for ( Match m = a.Match(result); m.Success; m = m.NextMatch() )
+                {
+                    String s = m.Groups[0].ToString();
+
+                    if ((s.StartsWith(":") ||
+                         s.StartsWith("@")) &&
+                         Parameters.Contains(s))
+                    {
+                        // It's a parameter. Lets handle it.
+
+                        NpgsqlParameter p = Parameters[s];
+                        if ((p.Direction == ParameterDirection.Input) ||
+                             (p.Direction == ParameterDirection.InputOutput))
+                        {
+
+                            // FIXME DEBUG ONLY
+                            // adding the '::<datatype>' on the end of a parameter is a highly
+                            // questionable practice, but it is great for debugging!
+                            sb.Append(p.TypeInfo.ConvertToBackend(p.Value, false));
+
+                            // Only add data type info if we are calling an stored procedure.
+
+                            if (type == CommandType.StoredProcedure)
+                            {
+                                sb.Append("::");
+                                sb.Append(p.TypeInfo.Name);
+
+                                if (p.TypeInfo.UseSize && (p.Size > 0))
+                                    sb.Append("(").Append(p.Size).Append(")");
+                            }
+                        }
+
+                    }
+                    else
+                        sb.Append(s);
+
+                }
+
+                result = sb.ToString();
+            }
+
+
+            else
+            {
+
+                for (Int32 i = 0; i < parameters.Count; i++)
+                {
+                    NpgsqlParameter Param = parameters[i];
+
+
+                    if ((Param.Direction == ParameterDirection.Input) ||
+                         (Param.Direction == ParameterDirection.InputOutput))
+
+
+                        result += Param.TypeInfo.ConvertToBackend(Param.Value, false) + "::" + Param.TypeInfo.Name + ",";
+                }
+
+
+                // Remove a trailing comma added from parameter handling above. If any.
+                // Maybe there are only output parameters. If so, there will be no comma.
+                if (result.EndsWith(","))
+                    result = result.Remove(result.Length - 1, 1);
+
+                result += ")";
+            }
+
+            if (functionReturnsRecord)
+                result = AddFunctionReturnsRecordSupport(result);
+
+            // If function returns ref cursor just process refcursor-result function call
+            // and return command which will be used to return data from refcursor.
+
+            if (functionReturnsRefcursor)
+                return ProcessRefcursorFunctionReturn(result);
+
+
+            return AddSingleRowBehaviorSupport(result);
+        }
+        
+        
+        
+        private Boolean CheckFunctionReturn(String ReturnType)
+        {
+
+            String returnRecordQuery = "select count(*) > 0 from pg_proc where prorettype = ( select oid from pg_type where typname = :typename ) and proargtypes=:proargtypes and proname=:proname;";
+
+            StringBuilder parameterTypes = new StringBuilder("");
+
+            foreach(NpgsqlParameter p in Parameters)
+            {
+                if ((p.Direction == ParameterDirection.Input) ||
+                     (p.Direction == ParameterDirection.InputOutput))
+                {
+                    parameterTypes.Append(Connection.Connector.OidToNameMapping[p.TypeInfo.Name].OID + " ");
+                }
+            }
+
+
+            NpgsqlCommand c = new NpgsqlCommand(returnRecordQuery, Connection);
+            
+            c.Parameters.Add(new NpgsqlParameter("typename", NpgsqlDbType.Text));
+            c.Parameters.Add(new NpgsqlParameter("proargtypes", NpgsqlDbType.Text));
+            c.Parameters.Add(new NpgsqlParameter("proname", NpgsqlDbType.Text));
+            
+            c.Parameters[0].Value = ReturnType;
+            c.Parameters[1].Value = parameterTypes.ToString();
+            c.Parameters[2].Value = CommandText;
+            
+
+            Boolean ret = (Boolean) c.ExecuteScalar();
+
+            // reset any responses just before getting new ones
+            connector.Mediator.ResetResponses();
+            return ret;
+
+
+        }
+        
+        
+        private String AddFunctionReturnsRecordSupport(String OriginalResult)
+        {
+                                
+            StringBuilder sb = new StringBuilder(OriginalResult);
+            
+            sb.Append(" as (");
+            
+            foreach(NpgsqlParameter p in Parameters)
+            {
+                if ((p.Direction == ParameterDirection.Output) ||
+                (p.Direction == ParameterDirection.InputOutput))
+                {
+                    sb.Append(String.Format("{0} {1}, ", p.ParameterName.Substring(1), p.TypeInfo.Name));
+                }
+            }
+            
+            String result = sb.ToString();
+            
+            result = result.Remove(result.Length - 2, 1);
+            
+            result += ")";
+            
+            
+            
+            return result;
+            
+            
+        }
+        
+        ///<summary>
+        /// This methods takes a string with a function call witch returns a refcursor or a set of
+        /// refcursor. It will return the names of the open cursors/portals which will hold
+        /// results. In turn, it returns the string which is needed to get the data of this cursors
+        /// in form of one resultset for each cursor open. This way, clients don't need to do anything
+        /// else besides calling function normally to get results in this way.
+        ///</summary>
+               
+        private String ProcessRefcursorFunctionReturn(String FunctionCall)
+        {
+            NpgsqlCommand c = new NpgsqlCommand(FunctionCall, Connection);
+            
+            NpgsqlDataReader dr = c.ExecuteReader();
+            
+            StringBuilder sb = new StringBuilder();
+            
+            while (dr.Read())
+            {
+                sb.Append("fetch all from \"").Append(dr.GetString(0)).Append("\";");
+                
+            }
+            
+            sb.Append(";"); // Just in case there is no response from refcursor function return.
+            
+            // reset any responses just before getting new ones
+            connector.Mediator.ResetResponses();
+            
+            return sb.ToString();
+                    
+            
         }
 
 
@@ -634,7 +1031,7 @@ namespace Npgsql
 
             for (Int32 i = 0; i < parameters.Count; i++)
             {
-                result.Append(NpgsqlTypesHelper.ConvertNpgsqlParameterToBackendStringValue(parameters[i], false) + ',');
+                result.Append(parameters[i].TypeInfo.ConvertToBackend(parameters[i].Value, false) + ',');
             }
 
             result = result.Remove(result.Length - 1, 1);
@@ -650,10 +1047,21 @@ namespace Npgsql
         {
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "GetParseCommandText");
 
+            Boolean addProcedureParenthesis = false;  // Do not add procedure parenthesis by default.
+            
             String parseCommand = text;
 
             if (type == CommandType.StoredProcedure)
+            {
+                // Check if just procedure name was passed. If so, does not replace parameter names and just pass parameter values in order they were added in parameters collection.
+                if (!parseCommand.Trim().EndsWith(")"))  
+                {
+                    addProcedureParenthesis = true;
+                    parseCommand += "(";
+                }
+                
                 parseCommand = "select * from " + parseCommand; // This syntax is only available in 7.3+ as well SupportsPrepare.
+            }
             else if (type == CommandType.TableDirect)
                 return "select * from " + parseCommand; // There is no parameter support on TableDirect.
 
@@ -666,15 +1074,28 @@ namespace Npgsql
 
                 for (i = 0; i < parameters.Count; i++)
                 {
-                    //result = result.Replace(":" + parameterName, parameters[i].Value.ToString());
-                    parameterName = parameters[i].ParameterName;
-                    //textCommand = textCommand.Replace(':' + parameterName, "$" + (i+1));
-                    parseCommand = ReplaceParameterValue(parseCommand, parameterName, "$" + (i+1));
+                    if ((parameters[i].Direction == ParameterDirection.Input) ||
+                    (parameters[i].Direction == ParameterDirection.InputOutput))
+                    {
+                    
+                        if (!addProcedureParenthesis)
+                        {
+                            //result = result.Replace(":" + parameterName, parameters[i].Value.ToString());
+                            parameterName = parameters[i].ParameterName;
+                            //textCommand = textCommand.Replace(':' + parameterName, "$" + (i+1));
+                            parseCommand = ReplaceParameterValue(parseCommand, parameterName, "$" + (i+1) + "::" + parameters[i].TypeInfo.Name);
+                        }
+                        else
+                            parseCommand += "$" + (i+1) + "::" + parameters[i].TypeInfo.Name;
+                    }
 
                 }
             }
 
-            return parseCommand;
+            if (addProcedureParenthesis)
+                return parseCommand + ")";
+            else
+                return parseCommand;
 
         }
 
@@ -683,16 +1104,25 @@ namespace Npgsql
         {
             NpgsqlEventLog.LogMethodEnter(LogLevel.Debug, CLASSNAME, "GetPrepareCommandText");
 
+            Boolean addProcedureParenthesis = false;  // Do not add procedure parenthesis by default.
 
-
-            planName = "NpgsqlPlan" + System.Threading.Interlocked.Increment(ref planIndex);
+            planName = Connector.NextPlanName();
 
             StringBuilder command = new StringBuilder("prepare " + planName);
 
             String textCommand = text;
 
             if (type == CommandType.StoredProcedure)
+            {
+                // Check if just procedure name was passed. If so, does not replace parameter names and just pass parameter values in order they were added in parameters collection.
+                if (!textCommand.Trim().EndsWith(")"))  
+                {
+                    addProcedureParenthesis = true;
+                    textCommand += "(";
+                }
+                
                 textCommand = "select * from " + textCommand;
+            }
             else if (type == CommandType.TableDirect)
                 return "select * from " + textCommand; // There is no parameter support on TableDirect.
 
@@ -706,12 +1136,22 @@ namespace Npgsql
 
                 for (i = 0; i < parameters.Count; i++)
                 {
-                    //result = result.Replace(":" + parameterName, parameters[i].Value.ToString());
-                    parameterName = parameters[i].ParameterName;
-                    // The space in front of '$' fixes a parsing problem in 7.3 server
-                    // which gives errors of operator when finding the caracters '=$' in
-                    // prepare text
-                    textCommand = ReplaceParameterValue(textCommand, parameterName, " $" + (i+1));
+                    if ((parameters[i].Direction == ParameterDirection.Input) ||
+                    (parameters[i].Direction == ParameterDirection.InputOutput))
+                    {
+                    
+                        if (!addProcedureParenthesis)
+                        {
+                            //result = result.Replace(":" + parameterName, parameters[i].Value.ToString());
+                            parameterName = parameters[i].ParameterName;
+                            // The space in front of '$' fixes a parsing problem in 7.3 server
+                            // which gives errors of operator when finding the caracters '=$' in
+                            // prepare text
+                            textCommand = ReplaceParameterValue(textCommand, parameterName, " $" + (i+1));
+                        }
+                        else
+                            textCommand += " $" + (i+1);
+                    }
 
                 }
 
@@ -722,7 +1162,8 @@ namespace Npgsql
 
                 for (i = 0; i < parameters.Count; i++)
                 {
-                    command.Append(NpgsqlTypesHelper.GetBackendTypeNameFromDbType(parameters[i].DbType));
+                    //                    command.Append(NpgsqlTypesHelper.GetDefaultTypeInfo(parameters[i].DbType));
+                    command.Append(parameters[i].TypeInfo.Name);
 
                     command.Append(',');
                 }
@@ -731,6 +1172,9 @@ namespace Npgsql
                 command.Append(')');
 
             }
+            
+            if (addProcedureParenthesis)
+                textCommand += ")";
 
             command.Append(" as ");
             command.Append(textCommand);
@@ -749,13 +1193,10 @@ namespace Npgsql
             Int32 paramEnd = paramStart + paramLen;
             Boolean found = false;
 
+
             while(paramStart > -1)
             {
-                if((resLen > paramEnd) &&
-                        (result[paramEnd] == ' ' ||
-                         result[paramEnd] == ',' ||
-                         result[paramEnd] == ')' ||
-                         result[paramEnd] == ';'))
+                if((resLen > paramEnd) && !Char.IsLetterOrDigit(result, paramEnd))
                 {
                     result = result.Substring(0, paramStart) + paramVal + result.Substring(paramEnd);
                     found = true;
@@ -775,8 +1216,29 @@ namespace Npgsql
             if(!found)
                 throw new IndexOutOfRangeException (String.Format(resman.GetString("Exception_ParamNotInQuery"), parameterName));
 
+
             return result;
         }//ReplaceParameterValue
+        
+        
+        private String AddSingleRowBehaviorSupport(String ResultCommandText)
+        {
+            
+            ResultCommandText = ResultCommandText.Trim();
+        
+            if ((commandBehavior & CommandBehavior.SingleRow) > 0)
+            {
+                if (ResultCommandText.EndsWith(";"))
+                    ResultCommandText = ResultCommandText.Substring(0, ResultCommandText.Length - 1);
+                ResultCommandText += " limit 1;";
+                
+            }
+            
+            
+            
+            return ResultCommandText;
+            
+        }
 
 
         private void ExecuteCommand()
@@ -784,27 +1246,44 @@ namespace Npgsql
             // Check the connection state first.
             CheckConnectionState();
 
-            if (parse == null) {
-                connection.Query(this);
+            // reset any responses just before getting new ones
+            connector.Mediator.ResetResponses();
+
+
+            if (parse == null)
+            {
+                Connector.Query(this);
 
                 // Check for errors and/or notifications and do the Right Thing.
-                connection.CheckErrorsAndNotifications();
-            } else {
-                BindParameters();
+                connector.CheckErrorsAndNotifications();
+            }
+            else
+            {
+                try
+                {
 
-                // Check for errors and/or notifications and do the Right Thing.
-                connection.CheckErrorsAndNotifications();
+                    BindParameters();
 
-                connection.Execute(new NpgsqlExecute(bind.PortalName, 0));
+                    connector.Execute(new NpgsqlExecute(bind.PortalName, 0));
 
-                // Check for errors and/or notifications and do the Right Thing.
-                connection.CheckErrorsAndNotifications();
+                    // Check for errors and/or notifications and do the Right Thing.
+                    connector.CheckErrorsAndNotifications();
+                }
+                finally
+                {
+                    // As per documentation:
+                    // "[...] When an error is detected while processing any extended-query message,
+                    // the backend issues ErrorResponse, then reads and discards messages until a
+                    // Sync is reached, then issues ReadyForQuery and returns to normal message processing.[...]"
+                    // So, send a sync command if we get any problems.
+
+                    connector.Sync();
+                }
             }
-            /*	else
-            		throw new NotImplementedException(resman.GetString("Exception_CommandTypeTableDirect"));*/
         }
-
-
+        
+        
+         
+        
     }
-
 }
